# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'lensql'
copyright = '2025, Davide Ponzini'
author = 'Davide Ponzini'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'autoapi.extension',
    # 'sphinx.ext.inheritance_diagram',
    # 'autoapi.sphinx',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']


# -- Autoapi -----------------------------------------------------------------
autoapi_dirs = ['../src']
autodoc_typehints = 'description'
autoapi_options = [ 'members', 'private-members', 'show-inheritance', 'show-module-summary', 'special-members', 'imported-members', ]



from . import html

from .. import server
from ..sql_errors import SQLException

from enum import Enum
from IPython.display import display, HTML
import ipywidgets as widgets
import pandas as pd
from dav_tools.chatgpt import MessageRole

CHAT_ID = 0
MSG_ID = 0

class Buttons(Enum):
    pass
    # SUCCESS_WRONG_OUTPUT = 'The output is not what I expected'
    # MANUAL_PROMPT = 'Other'

class ResultButtons(Buttons):
    DESCRIBE = 'Describe query'
    EXPLAIN = 'Explain query'

class ErrorButtons(Buttons):
    EXPLAIN = 'Explain error'
    EXAMPLE = 'Show example'
    LOCATE = 'Where to look'
    FIX = 'Suggest fix'


class Message:
    '''Represents a message in the chat.'''
    def __init__(self, role: MessageRole, content: str):
        global MSG_ID
        MSG_ID += 1
        self.msg_id = MSG_ID
        self.role = role
        self.content = content
        self.html = html.Message(role, content, self.msg_id)


class Chat:
    '''Generic chat for displaying messages and interacting with the user.'''
    def __init__(self):
        global CHAT_ID
        CHAT_ID += 1
        self.chat_id = CHAT_ID
        self.messages: list[Message] = []
        self.thinking_msg = None

        self.output_widget = widgets.Output()  # Capture output in Jupyter cell

    @property
    def last_message_id(self) -> int:
        if len(self.messages) == 0:
            return -1
        return self.messages[-1].msg_id

    def show(self) -> None:
        display(self.output_widget)  # Ensure output appears inside the correct cell

        chat_html = html.Chat(self.chat_id)
        self.display_html(chat_html)

    def display_html(self, content: str):
        with self.output_widget:
            display(HTML(str(content)))

    def display_box(self, content: list[widgets.Widget]):
        with self.output_widget:
            display(widgets.HBox(content))

    def show_message(self, role: MessageRole, text: str) -> Message:
        message = Message(role, text)
        self.messages.append(message)

        message_html = str(message.html)
        message_html = message_html.replace('`', '\\`') #.replace('\n', '<br>')

        append_script = f'''
            <script>
                var target = document.getElementById('chat{self.chat_id}');
                target.insertAdjacentHTML('beforeend', `{message_html}`);
                target.scrollTop = target.scrollHeight;
            </script>
        '''

        self.display_html(append_script)
        return message
    
    def delete_message(self, msg_id: int) -> None:
        delete_script = f'''
            <script>
                var target = document.getElementById('msg{msg_id}');
                target.remove();
            </script>
        '''
        self.display_html(delete_script)

    def is_thiking(self) -> bool:
        '''Returns whether the assistant is currently thinking.'''
        return self.thinking_msg is not None

    def start_thinking(self) -> None:
        '''Displays a `Thinking...` message in the chat.'''
        if self.is_thiking():
            self.stop_thinking()

        self.thinking_msg = self.show_message(MessageRole.ASSISTANT, 'Thinking...')

    def stop_thinking(self) -> None:
        '''Removes the `Thinking...` message from the chat.'''
        if not self.is_thiking():
            return
        
        self.delete_message(self.thinking_msg.msg_id)
        self.thinking_msg = None

    def get_input(self, cb):
        """Creates a text input field for the user to type in."""
        text_input = widgets.Text(placeholder='Type here...', layout=widgets.Layout(width='100%'))
        send_button = widgets.Button(description='Send')

        # Define the on_submit event
        def on_submit(b):
            user_text = text_input.value.strip()
            if user_text:
                cb(user_text)
                text_input.close()
                send_button.close()
        # -- End of on_submit event --

        text_input.on_submit(on_submit)
        send_button.on_click(on_submit)

        self.display_box([text_input, send_button])

class CodeChat(Chat):
    '''Chat related to a specific code snippet.'''
    def __init__(self, query_id : int, code: str, data):
        super().__init__()

        self.query_id = query_id
        self.code = code
        self.data = data


class ResultChat(CodeChat):
    '''Chat related to a specific code snippet that has returned a result.'''
    def __init__(self, query_id: int, code: str, result: pd.DataFrame):
        super().__init__(query_id, code, data=result)

    def show(self):
        super().show()

        self.show_buttons()

    def show_buttons(self):
        options = [option.value for option in ResultButtons]
        buttons = [widgets.Button(description=option) for option in options]

        buttons[0].layout.margin = '2px 2px 2px 62px'
        for button in buttons[1:]:
            button.layout.margin = '2px 2px 2px 2px'

        # Define the on_click event
        def on_button_click(b):
            button_text = b.description

            for button in buttons:
                button.close()

            self.show_message(MessageRole.USER, button_text)
            self.start_thinking()

            if button_text == ResultButtons.DESCRIBE.value:
                response = server.describe_my_query(self.query_id, self.chat_id, self.last_message_id)
            elif button_text == ResultButtons.EXPLAIN.value:
                response = server.explain_my_query(self.query_id, self.chat_id, self.last_message_id)
            else:
                response = 'Action not implemented yet.'

            self.stop_thinking()
            self.show_message(MessageRole.ASSISTANT, response)
            self.show_buttons()
        # -- End of on_button_click event --

        # Assign on_click event to each button
        for button in buttons:
            button.on_click(on_button_click)

        self.display_box(buttons)


class ErrorChat(CodeChat):
    '''Chat related to a specific code snippet that has returned an error.'''
    def __init__(self, query_id: int, code, exception: Exception):
        super().__init__(query_id, code, data=SQLException(exception))
        
    def show(self):
        super().show()

        message = html.exception_to_html(self.data)
        self.show_message(MessageRole.ASSISTANT, message)
        self.show_buttons()

    def show_buttons(self):
        """Creates a set of buttons for the user to choose from."""
        options = [option.value for option in ErrorButtons]
        buttons = [widgets.Button(description=option) for option in options]
        
        buttons[0].layout.margin = '2px 2px 2px 62px'
        for button in buttons[1:]:
            button.layout.margin = '2px 2px 2px 2px'

        # Define the on_click event
        def on_button_click(b):
            button_text = b.description

            for button in buttons:
                button.close()

            self.show_message(MessageRole.USER, button_text)
            self.start_thinking()

            if button_text == ErrorButtons.EXPLAIN.value:
                response = server.explain_error_message(self.query_id, str(self.data), self.chat_id, self.last_message_id)
            elif button_text == ErrorButtons.EXAMPLE.value:
                response = server.provide_error_example(self.query_id, str(self.data), self.chat_id, self.last_message_id)
            elif button_text == ErrorButtons.LOCATE.value:
                response = server.locate_error_cause(self.query_id, str(self.data), self.chat_id, self.last_message_id)
            elif button_text == ErrorButtons.FIX.value:
                response = server.fix_query(self.query_id, str(self.data), self.chat_id, self.last_message_id)
            else:
                response = 'Action not implemented yet.'

            self.stop_thinking()
            self.show_message(MessageRole.ASSISTANT, response)
            self.show_buttons()

        # -- End of on_button_click event --

        # Assign on_click event to each button
        for button in buttons:
            button.on_click(on_button_click)

        self.display_box(buttons)


import psycopg2
from psycopg2.extensions import connection
from dav_tools import messages

HOST = None
PORT = None
DBNAME = None
USERNAME = None
PASSWORD = None

def are_credentials_set():
    '''Checks if all database credentials are set.'''

    for cred in (HOST, PORT, DBNAME, USERNAME, PASSWORD):
        if cred is None:
            return False
    return True

def connect() -> connection:
    '''Opens a connection to the database.'''

    if not are_credentials_set():
        raise Exception('Database credentials not set')
    
    return psycopg2.connect(
        host=HOST,
        port=PORT,
        dbname=DBNAME,
        user=USERNAME,
        password=PASSWORD
    )

def set_credentials(host: str, port: int, dbname: str, username: str, password: str) -> bool:
    '''Sets the database credentials and tests the connection.'''
    
    global HOST, PORT, DBNAME, USERNAME, PASSWORD

    HOST = host
    PORT = port
    DBNAME = dbname
    USERNAME = username
    PASSWORD = password

    try:
        conn = connect()
        conn.close()

        return True
    except Exception as e:
        messages.error('Error connecting to the database:', e)
        return False


from . import load_content
from ..sql_errors import SQLException

from dav_tools.chatgpt import MessageRole

import html

CSS = load_content.load_from_this_dir('style.css')


class HtmlComponent:
    def __init__(self, html: str):
        self.html = html

    def __str__(self):
        return self.html


class Icon:
    USER = '''
        <div class="icon">
            <i class="fas fa-user"></i>
            <br>
            You 
        </div>
    '''
    ASSISTANT = '''
        <div class="icon">
            <i class="fas fa-search"></i>
            <br>
            LensQL
        </div>
    '''
    NO_ICON = ''

def exception_to_html(exception: SQLException) -> str:
    traceback = '\n' + '\n'.join(exception.traceback)
    traceback = html.escape(traceback)

    return f'''
        <b class="m">{exception}</b>
        <br>
        <pre class="code m">{traceback}</pre>
    '''

class Chat(HtmlComponent):
    def __init__(self, id: int):
        super().__init__(f'''
                <style>{CSS}</style>
            
                <div class="box" id="chat{id}"></div>
            ''')

class Message(HtmlComponent):
    def __init__(self, role: MessageRole, content: str, msg_id: int):
        super().__init__(f'''
                <div class="messagebox messagebox-{role}" id="msg{msg_id}">
                    {Icon.ASSISTANT if role == MessageRole.ASSISTANT else Icon.NO_ICON}    
                    <div class="message">
                        {content}
                    </div>
                    {Icon.USER if role == MessageRole.USER else Icon.NO_ICON}
                </div>
            ''')



class SQLException:
    def __init__(self, exception: Exception):
        self.exception = exception

        self.name = type(self.exception).__name__

        message = str(self.exception.args[0]) if self.exception.args else ''
        self.description = message.splitlines()[0]
        self.traceback = message.splitlines()[1:]
    
    def __str__(self):
        return f'{self.name}: {self.description}'


import requests
import json
from dav_tools import messages

HOST = None
USERNAME = None

def login(host: str, username: str) -> bool:
    '''Connects the user to the server.'''
    global HOST, USERNAME

    HOST = host
    USERNAME = username

    try:
        response = requests.post(f'{HOST}/login', data={
            'username': json.dumps(USERNAME)
        }).json()
        
        if response['status'] == 'ok':
            return True
        
        messages.error('Error connecting user to the server:', response['message'])
        return False
    except Exception as e:
        messages.error('Error connecting user to the server:', e)
        return False


def log_query(query: str, success: bool) -> int:
    response = requests.post(f'{HOST}/log-query', data={
        'username': json.dumps(USERNAME),
        'query': json.dumps(query),
        'success': json.dumps(success)
    }).json()

    return response['query_id']

def explain_error_message(query_id: int, exception: str, chat_id: int, msg_id: int) -> str:
    response = requests.post(f'{HOST}/explain-error-message', data={
        'username': json.dumps(USERNAME),
        'query_id': json.dumps(query_id),
        'exception': json.dumps(exception),
        'chat_id': json.dumps(chat_id),
        'msg_id': json.dumps(msg_id)
    }).json()

    return response['answer']

def locate_error_cause(query_id: int, exception: str, chat_id: int, msg_id: int) -> str:
    response = requests.post(f'{HOST}/locate-error-cause', data={
        'username': json.dumps(USERNAME),
        'query_id': json.dumps(query_id),
        'exception': json.dumps(exception),
        'chat_id': json.dumps(chat_id),
        'msg_id': json.dumps(msg_id)
    }).json()

    return response['answer']

def provide_error_example(query_id: int, exception:str, chat_id: int, msg_id: int) -> str:
    response = requests.post(f'{HOST}/provide-error-example', data={
        'username': json.dumps(USERNAME),
        'query_id': json.dumps(query_id),
        'exception': json.dumps(exception),
        'chat_id': json.dumps(chat_id),
        'msg_id': json.dumps(msg_id)
    }).json()

    return response['answer']

def fix_query(query_id: int, expection: str, chat_id: int, msg_id: int) -> str:
    response = requests.post(f'{HOST}/fix-query', data={
        'username': json.dumps(USERNAME),
        'query_id': json.dumps(query_id),
        'exception': json.dumps(expection),
        'chat_id': json.dumps(chat_id),
        'msg_id': json.dumps(msg_id)
    }).json()
    
    return response['answer']

def describe_my_query(query_id: int, chat_id: int, msg_id: int) -> str:
    response = requests.post(f'{HOST}/describe-my-query', data={
        'username': json.dumps(USERNAME),
        'query_id': json.dumps(query_id),
        'chat_id': json.dumps(chat_id),
        'msg_id': json.dumps(msg_id)
    }).json()
    
    return response['answer']

def explain_my_query(query_id: int, chat_id: int, msg_id: int) -> str:
    response = requests.post(f'{HOST}/explain-my-query', data={
        'username': json.dumps(USERNAME),
        'query_id': json.dumps(query_id),
        'chat_id': json.dumps(chat_id),
        'msg_id': json.dumps(msg_id)
    }).json()
    
    return response['answer']



from . import chat

from .. import server

import pandas as pd
from IPython.display import display

def show_result(code: str, result: pd.DataFrame) -> None:
    display(result)

    query_id = server.log_query(query=code, success=True)
    chat.ResultChat(query_id, code, result).show()

def show_error(code: str, exception: Exception) -> None:
    query_id = server.log_query(query=code, success=False)
    chat.ErrorChat(query_id, code, exception).show()



import os
import base64

_cache = {}

def load_from_this_dir(file):
    file = f'{os.path.dirname(os.path.abspath(__file__))}/{file}'
    try:
        if file.endswith('.png'):
            file_data = open(file, "rb").read()
            return 'data:image/png;base64,' + base64.b64encode(file_data).decode('utf-8')
        else:
            with open(file) as f:
                return f.read()
    except:
        print('Cannot load', file)
        return ''


def load_icon(file):
    if file.startswith('https://') or file.startswith('http://'):
        return file
    if file not in _cache:
        _cache[file] = load_from_this_dir(file)
    return _cache[file]


def load_css(file):
    assert file.endswith('.css')
    if file not in _cache:
        _cache[file] = load_from_this_dir(file)
    data = _cache[file]
    return f'<style>{data}</style>'


def load_js(file):
    assert file.endswith('.js')
    if file not in _cache:
        _cache[file] = load_from_this_dir(file)
    data = _cache[file]
    return f'<script>{data}</script>'



from .. import webui

from .. import database

from dav_tools import messages
import psycopg2
from IPython.core.interactiveshell import ExecutionResult, ExecutionInfo
import pandas as pd

SQL_COMMANDS = ('SELECT', 'INSERT', 'UPDATE', 'DELETE', 'CREATE', 'DROP', 'ALTER', 'SET')


def run_cell_sql_python(shell, raw_cell: str, **kwargs) -> ExecutionResult:
    '''Executes SQL queries when the cell starts with a SQL command. Otherwise, executes Python code.'''

    if not raw_cell.strip().upper().startswith(SQL_COMMANDS):
        return shell.run_cell_original(raw_cell, **kwargs)
    
    return run_cell_sql_only(shell, raw_cell, **kwargs)

def run_cell_sql_only(shell, raw_cell: str, **kwargs) -> ExecutionResult:
    '''Executes SQL queries.'''

    if not database.are_credentials_set():
        messages.error('Module not configured, please run the setup function first')
        return

    # execute SQL query
    try:
        conn = database.connect()
        cur = conn.cursor()
        cur.execute(raw_cell)
        conn.commit()
        
        if cur.description:  # Check if the query has a result set
            rows = cur.fetchall()
            columns = [desc[0] for desc in cur.description]
            result = pd.DataFrame(rows, columns=columns)
            
            webui.show_result(raw_cell, result)
        else:
            print(f'Affected rows: {cur.rowcount}')
            result = None

        return return_result(result, raw_cell, **kwargs)
    except Exception as e:
        webui.show_error(code=raw_cell, exception=e)
        return
    finally:
        cur.close()
        conn.close()

def return_result(result, raw_cell, store_history, silent, cell_id, shell_futures=True) -> ExecutionResult:
    res = ExecutionResult(ExecutionInfo(raw_cell, store_history, silent, shell_futures, cell_id))
    res.result = result

    return res




'''
This module contains the functions to configure the database connection and enable SQL execution in the notebook.
'''

from . import run_cell

from .. import database
from .. import server

from IPython.core.interactiveshell import InteractiveShell
from dav_tools import messages
import sys
import pandas as pd
import os

def setup(
        host: str = '', username: str = '',
        dbhost: str     =       os.getenv('LENSQL_DB_HOST',     ''),
        dbport: int     = int(  os.getenv('LENSQL_DB_PORT',     '5432')),
        dbname: str     =       os.getenv('LENSQL_DB_NAME',     ''),
        dbusername: str =       os.getenv('LENSQL_DB_USERNAME', ''),
        dbpassword: str =       os.getenv('LENSQL_DB_PASSWORD', ''),
        *, allow_code_execution=False):
    '''Configures the database connection and enables SQL execution in the notebook.'''
    while len(host) == 0: 
        host = messages.ask('Enter server host', file=sys.stdout)
    while len(username) == 0:
        username = messages.ask('Enter server username', file=sys.stdout)

    if server.login(host, username):
        messages.success('Server is online', file=sys.stdout)
    else:
        return
    
    while len(dbhost) == 0:
        dbhost = messages.ask('Enter database host', file=sys.stdout)
    while dbport is None:
        dbport = messages.ask('Enter database port', file=sys.stdout)
    while len(dbname) == 0:
        dbname = messages.ask('Enter database name', file=sys.stdout)
    while len(dbusername) == 0:
        dbusername = messages.ask('Enter database username', file=sys.stdout)
    while len(dbpassword) == 0:
        dbpassword = messages.ask('Enter database password', secret=True, file=sys.stdout)

    if database.set_credentials(dbhost, dbport, dbname, dbusername, dbpassword):
        messages.success('Connected to database', file=sys.stdout)
    else:
        return

    # Display all DataFrame rows
    pd.set_option('display.max_rows', None)

    override_execution(allow_code_execution)


def override_execution(allow_code_execution=False):
    if allow_code_execution and not hasattr(InteractiveShell, 'run_cell_original'):
        InteractiveShell.run_cell_original = InteractiveShell.run_cell
        InteractiveShell.run_cell = run_cell.run_cell_sql_python
    else:
        InteractiveShell.run_cell = run_cell.run_cell_sql_only

    messages.success('SQL execution enabled', file=sys.stdout)
    if allow_code_execution:
        messages.warning(f'Only commands starting with {run_cell.SQL_COMMANDS} will be interpreted as SQL', file=sys.stdout)
    else:
        messages.info('All code executed from now on will be interpreted as SQL', file=sys.stdout)
